@@ -495,6 +495,7 @@ impl<'a> SerializedModule<'a> {
             interruptable,
             consume_fuel,
             static_memory_bound_is_maximum,
+            guard_before_linear_memory,
         } = self.tunables;
 
         let other = compiler.tunables();
@@ -531,6 +532,11 @@ impl<'a> SerializedModule<'a> {
             other.static_memory_bound_is_maximum,
             "pooling allocation support",
         )?;
+        Self::check_bool(
+            guard_before_linear_memory,
+            other.guard_before_linear_memory,
+            "guard before linear memory",
+        )?;
 
         Ok(())
     }
@@ -667,7 +673,6 @@ mod test {
         Ok(())
     }
 
-    #[cfg(target_arch = "x86_64")]
     #[test]
     fn test_isa_flags_mismatch() -> Result<()> {
         let engine = Engine::default();
