@@ -36,7 +36,7 @@ pub use self::data::*;
 /// [`Store`] it will not be deallocated until the [`Store`] itself is dropped.
 /// This makes [`Store`] unsuitable for creating an unbounded number of
 /// instances in it because [`Store`] will never release this memory. It's
-/// recommended to have a [`Store`] correspond roughly to the liftime of a "main
+/// recommended to have a [`Store`] correspond roughly to the lifetime of a "main
 /// instance" that an embedding is interested in executing.
 ///
 /// ## Type parameter `T`
@@ -73,7 +73,7 @@ pub use self::data::*;
 ///
 /// You can create a store with default configuration settings using
 /// `Store::default()`. This will create a brand new [`Engine`] with default
-/// ocnfiguration (see [`Config`](crate::Config) for more information).
+/// configuration (see [`Config`](crate::Config) for more information).
 pub struct Store<T> {
     // for comments about `ManuallyDrop`, see `Store::into_data`
     inner: ManuallyDrop<Box<StoreInner<T>>>,
@@ -178,7 +178,7 @@ struct StoreInstance {
 enum OutOfGas {
     Trap,
     InjectFuel {
-        injection_count: u32,
+        injection_count: u64,
         fuel_to_inject: u64,
     },
 }
@@ -327,10 +327,25 @@ impl<T> Store<T> {
         inner.limiter = Some(Box::new(limiter));
     }
 
-    /// Configure a function that runs each time WebAssembly code running on this [`Store`] calls
-    /// into native code.
+    /// Configure a function that runs each time the host resumes execution from
+    /// WebAssembly.
     ///
-    /// This function may return a [`Trap`], which terminates execution.
+    /// This hook is called in two circumstances:
+    ///
+    /// * When WebAssembly calls a function defined by the host, this hook is
+    ///   called before other host code runs.
+    /// * When WebAssembly returns back to the host after being called, this
+    ///   hook is called.
+    ///
+    /// This method can be used with [`Store::exiting_native_code_hook`] to track
+    /// execution time of WebAssembly, for example, by starting/stopping timers
+    /// in the enter/exit hooks.
+    ///
+    /// This function may return a [`Trap`]. If a trap is returned when an
+    /// import was called, it is immediately raised as-if the host import had
+    /// returned the trap. If a trap is returned after wasm returns to the host
+    /// then the wasm function's result is ignored and this trap is returned
+    /// instead.
     pub fn entering_native_code_hook(
         &mut self,
         hook: impl FnMut(&mut T) -> Result<(), Trap> + Send + Sync + 'static,
@@ -338,10 +353,25 @@ impl<T> Store<T> {
         self.inner.entering_native_hook = Some(Box::new(hook));
     }
 
-    /// Configure a function that runs before native code running on this [`Store`] returns to
-    /// WebAssembly code.
+    /// Configure a function that runs just before WebAssembly code starts
+    /// executing.
     ///
-    /// This function may return a [`Trap`], which terminates execution.
+    /// The closure provided is called in two circumstances:
+    ///
+    /// * When the host calls a WebAssembly function, the hook is called just
+    ///   before WebAssembly starts executing.
+    /// * When a host function returns back to WebAssembly this hook is called
+    ///   just before the return.
+    ///
+    /// This method can be used with [`Store::entering_native_code_hook`] to track
+    /// execution time of WebAssembly, for example, by starting/stopping timers
+    /// in the enter/exit hooks.
+    ///
+    /// This function may return a [`Trap`]. If a trap is returned when an
+    /// imported host function is returning, then the imported host function's
+    /// result is ignored and the trap is raised. If a trap is returned when
+    /// the host is about to start executing WebAssembly, then no WebAssembly
+    /// code is run and the trap is returned instead.
     pub fn exiting_native_code_hook(
         &mut self,
         hook: impl FnMut(&mut T) -> Result<(), Trap> + Send + Sync + 'static,
@@ -528,7 +558,7 @@ impl<T> Store<T> {
     ///
     /// This method will panic if it is not called on a store associated with an [async
     /// config](crate::Config::async_support).
-    pub fn out_of_fuel_async_yield(&mut self, injection_count: u32, fuel_to_inject: u64) {
+    pub fn out_of_fuel_async_yield(&mut self, injection_count: u64, fuel_to_inject: u64) {
         self.inner
             .out_of_fuel_async_yield(injection_count, fuel_to_inject)
     }
@@ -625,14 +655,10 @@ impl<'a, T> StoreContextMut<'a, T> {
     /// runs out.
     ///
     /// For more information see [`Store::out_of_fuel_async_yield`]
-    pub fn out_of_fuel_async_yield(&mut self, injection_count: u32, fuel_to_inject: u64) {
+    pub fn out_of_fuel_async_yield(&mut self, injection_count: u64, fuel_to_inject: u64) {
         self.0
             .out_of_fuel_async_yield(injection_count, fuel_to_inject)
     }
-
-    pub(crate) fn store_data(self) -> &'a StoreData {
-        self.0.store_data()
-    }
 }
 
 impl<T> StoreInner<T> {
@@ -812,7 +838,7 @@ impl StoreInnermost {
         self.out_of_gas_behavior = OutOfGas::Trap;
     }
 
-    fn out_of_fuel_async_yield(&mut self, injection_count: u32, fuel_to_inject: u64) {
+    fn out_of_fuel_async_yield(&mut self, injection_count: u64, fuel_to_inject: u64) {
         assert!(
             self.async_support(),
             "cannot use `out_of_fuel_async_yield` without enabling async support in the config"
@@ -925,7 +951,7 @@ impl StoreInnermost {
     }
 }
 
-impl StoreOpaqueSend<'_> {
+impl<T> StoreContextMut<'_, T> {
     /// Executes a synchronous computation `func` asynchronously on a new fiber.
     ///
     /// This function will convert the synchronous `func` into an asynchronous
@@ -936,21 +962,23 @@ impl StoreOpaqueSend<'_> {
     /// necessary to suspend the fiber later on and poll sub-futures. It's hoped
     /// that the various comments are illuminating as to what's going on here.
     #[cfg(feature = "async")]
-    pub async fn on_fiber<R>(
+    pub(crate) async fn on_fiber<R>(
         &mut self,
-        func: impl FnOnce(&mut StoreOpaque<'_>) -> R + Send,
-    ) -> Result<R, Trap> {
-        let config = self.engine.config();
-
-        debug_assert!(self.async_support());
+        func: impl FnOnce(&mut StoreContextMut<'_, T>) -> R + Send,
+    ) -> Result<R, Trap>
+    where
+        T: Send,
+    {
+        let config = self.engine().config();
+        debug_assert!(self.0.async_support());
         debug_assert!(config.async_stack_size > 0);
 
         let mut slot = None;
         let future = {
-            let current_poll_cx = self.async_state.current_poll_cx.get();
-            let current_suspend = self.async_state.current_suspend.get();
+            let current_poll_cx = self.0.async_state.current_poll_cx.get();
+            let current_suspend = self.0.async_state.current_suspend.get();
             let stack = self
-                .engine
+                .engine()
                 .allocator()
                 .allocate_fiber_stack()
                 .map_err(|e| Trap::from(anyhow::Error::from(e)))?;
@@ -974,7 +1002,7 @@ impl StoreOpaqueSend<'_> {
                     let _reset = Reset(current_suspend, *current_suspend);
                     *current_suspend = suspend;
 
-                    *slot = Some(func(&mut self.opaque()));
+                    *slot = Some(func(self));
                     Ok(())
                 }
             })
