@@ -42,6 +42,14 @@ impl fmt::Display for TrapReason {
 /// A trap code describing the reason for a trap.
 ///
 /// All trap instructions have an explicit trap code.
+///
+/// The code can be accessed from the c-api, where the possible values are translated
+/// into enum values defined there:
+///
+/// * `wasm_trap_code` in c-api/src/trap.rs, and
+/// * `wasmtime_trap_code_enum` in c-api/include/wasmtime/trap.h.
+///
+/// These need to be kept in sync.
 #[non_exhaustive]
 #[derive(Clone, Copy, PartialEq, Eq, Debug, Hash)]
 pub enum TrapCode {
@@ -264,6 +272,16 @@ impl Trap {
         }
     }
 
+    /// Displays the error reason for this trap.
+    ///
+    /// In particular, it differs from this struct's `Display` by *only*
+    /// showing the reason, and not the full backtrace. This is useful to
+    /// customize the way the trap is reported, for instance to display a short
+    /// message for user-facing errors.
+    pub fn display_reason<'a>(&'a self) -> impl fmt::Display + 'a {
+        &self.inner.reason
+    }
+
     /// Returns a list of function frames in WebAssembly code that led to this
     /// trap happening.
     pub fn trace(&self) -> &[FrameInfo] {
@@ -345,7 +363,7 @@ impl fmt::Display for Trap {
             }
         }
         if self.inner.hint_wasm_backtrace_details_env {
-            writeln!(f, "note: run with `WASMTIME_BACKTRACE_DETAILS=1` environment variable to display more information")?;
+            writeln!(f, "note: using the `WASMTIME_BACKTRACE_DETAILS=1` environment variable to may show more debugging information")?;
         }
         Ok(())
     }
